from __future__ import annotations
import asyncio
import jinja2
import asyncio
from aiofauna import asyncio
from aiohttp_sse import sse_response
from aiohttp.web import (
    Application,
    AppRunner,
    TCPSite,
    Request,
    Response,
    json_response,
)
from aiohttp import ClientSession
from typing import Callable, Coroutine
from functools import wraps
from inspect import signature
from aiofauna.odm import AsyncFaunaModel
from pydantic import BaseModel
from aiohttp.web_request import FileField
from aiohttp.web import (
    Application,
    AppRunner,
    TCPSite,
    Request,
    Response,
    json_response,
)


class App(Application):
    """
    
    This module defines the App class, which is a subclass of the aiohttp.web.Application class.
    The App class provides a simple way to create web applications with Jinja2 templates.

    Usage Example:
    
    ```python
    
    from aiohttp import web

    app = App()

    @app.template("index.html", func=example_function)
    async def index(request):
        return {"message": "Hello, World!"}

    @app.route("GET", "/hello")
    async def hello(request):
        return web.Response(text="Hello, World!")

    web.run_app(app)

     Args:
        *args: Positional arguments to be passed to the aiohttp.web.Application superclass.
        **kwargs: Keyword arguments to be passed to the aiohttp.web.Application superclass.
        
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader("templates"))
        self.openapi = {
            "openapi": "3.0.0",
            "info": {"title": "Aiofauna", "version": "0.1.0"},
            "paths": {},
            "components": {},
        }

    def template(self, template_name: str, **kwargs):
        """
        Decorator to render a template and return an HTTP response.

        Args:
            template_name (str): The name of the template file.
            **kwargs: Keyword arguments representing context variables to be passed to the template.

        Returns:
            Callable: The decorator function.
        """

        def decorator(func):
            @wraps(func)
            async def wrapper(request: Request) -> Response:
                if asyncio.iscoroutinefunction(func):
                    data = await func(request)
                else:
                    data = func(request)
                template = self.jinja_env.get_template(template_name).render(**data)
                return Response(body=template.encode(), content_type="text/html")

            return wrapper

        return decorator

    def document(self, method: str, path: str, **kwargs):
        """
        Inspects the signature of the handler function and based on the parameters it constraints the Request object to the parameters shape.
        
        The parameters can be aither `Scalars` or `Models`.
        
        `Scalars`: `str`, `int`, `float`, `bool`
        
        `Models`: `BaseModel`, `AsyncFaunaModel`, `FileField`
        
        Those parameters are enforced and bound on the Request object. And added to the openapi documentation property.
        

        
        Args:
            method (str): The HTTP method for the route (e.g., "GET", "POST").
            path (str): The URL path for the route.
            **kwargs: Additional keyword arguments for the route.
            
        Returns:
        
            Callable: The decorator function.
            
        """

        def decorator(func):
            async def wrapper(request) -> Response:
                sig = signature(func)
                params = sig.parameters
                args_to_apply = {}
                open_api_params = {}
                for k, v in params.items():
                    if v.annotation in [str, int, float, bool]:
                        if k in request.query:
                            request[k] = v.annotation(request.query[k])
                            open_api_params[k] = {
                                "in": "query",
                                "name": k,
                                "schema": {"type": v.annotation.__name__},
                            }
                            args_to_apply[k] = request[k]
                        elif k in request.match_info:
                            request[k] = v.annotation(request.match_info[k])
                            open_api_params[k] = {
                                "in": "path",
                                "name": k,
                                "schema": {"type": v.annotation.__name__},
                            }
                            args_to_apply[k] = request[k]
                        elif k in request.headers:
                            request[k] = v.annotation(request.headers[k])
                            open_api_params[k] = {
                                "in": "header",
                                "name": k,
                                "schema": {"type": v.annotation.__name__},
                            }
                            args_to_apply[k] = request[k]
                        elif k in request.cookies:
                            request[k] = v.annotation(request.cookies[k])
                            open_api_params[k] = {
                                "in": "cookie",
                                "name": k,
                                "schema": {"type": v.annotation.__name__},
                            }
                            args_to_apply[k] = request[k]
                        else:
                            raise Exception(f"Missing parameter {k}")
                    elif v.annotation.__class__ in [BaseModel, AsyncFaunaModel]:
                        if request.method not in ["POST", "PUT", "PATCH"]:
                            raise Exception(
                                f"Method {request.method} not allowed for {k}"
                            )
                        if request.content_type == "application/json":
                            open_api_params[k] = {
                                "in": "body",
                                "name": k,
                                "schema": {
                                    "$ref": f"#/components/schemas/{v.annotation.__name__}"
                                },
                            }
                            request[k] = v.annotation(**await request.json())
                            args_to_apply[k] = request[k]
                        elif request.content_type == "multipart/form-data":
                            open_api_params[k] = {
                                "in": "body",
                                "name": k,
                                "schema": {
                                    "$ref": f"#/components/schemas/{v.annotation.__name__}"
                                },
                            }
                            request[k] = v.annotation(**await request.post())
                            args_to_apply[k] = request[k]
                        else:
                            raise Exception(
                                f"Content type {request.content_type} not supported"
                            )
                    elif v.annotation.__class__ == FileField:
                        if request.method not in ["POST", "PUT", "PATCH"]:
                            raise Exception(
                                f"Method {request.method} not allowed for {k}"
                            )
                        if request.content_type == "multipart/form-data":
                            open_api_params[k] = {
                                "in": "body",
                                "name": k,
                                "schema": {"type": "string", "format": "binary"},
                            }
                            request[k] = await request.post()
                            args_to_apply[k] = request[k]
                        else:
                            raise Exception(
                                f"Content type {request.content_type} not supported"
                            )
                    else:
                        raise Exception(f"Type {v.annotation} not supported")
                    self.openapi["paths"][path] = {
                        method.lower(): {
                            "summary": func.__name__,
                            "description": func.__doc__,
                            "parameters": list(open_api_params.values()),
                            "responses": {"200": {"description": "OK"}},
                        }
                    }
                if asyncio.iscoroutinefunction(func):
                    return await func(request, **args_to_apply)
                return func(request, **args_to_apply)

            return wrapper

        return decorator

    async def listen(self, host: bool = False, port: int = 8000, **kwargs):

        """
        
        Starts Listening for requests.
        
        Args:
        
            host (str): The host to listen on.
            
            port (int): The port to listen on.
            
            **kwargs: Additional keyword arguments for the server.
            
        """
        if not host:
            _host = "localhost"
        _host = "0.0.0.0"
        runner = AppRunner(self, **kwargs)
        await runner.setup()
        site = TCPSite(runner, _host, port)
        await site.start()
        while True:
            await asyncio.sleep(1)

    def get(self, path: str, **kwargs):
        """
        Decorator to add a GET route.

        Args:
            path (str): The URL path for the route.
            **kwargs: Additional keyword arguments for the route.

        Returns:
            Callable: The decorator function.
        """

        @wraps(self.document("GET", path, **kwargs))
        def decorator(func):
            self.router.add_route(
                "GET", path, self.document("GET", path, **kwargs)(func), **kwargs
            )
            return func

        return decorator

    def post(self, path: str, **kwargs):
        """
        Decorator to add a POST route.

        Args:
            path (str): The URL path for the route.
            **kwargs: Additional keyword arguments for the route.

        Returns:
            Callable: The decorator function.
        """

        @wraps(self.document("POST", path, **kwargs))
        def decorator(func):
            self.router.add_route(
                "POST", path, self.document("POST", path, **kwargs)(func), **kwargs
            )
            return func

        return decorator

    def put(self, path: str, **kwargs):
        """
        Decorator to add a PUT route.

        Args:
            path (str): The URL path for the route.
            **kwargs: Additional keyword arguments for the route.

        Returns:
            Callable: The decorator function.
        """

        @wraps(self.document("PUT", path, **kwargs))
        def decorator(func):
            self.router.add_route(
                "PUT", path, self.document("PUT", path, **kwargs)(func), **kwargs
            )
            return func

        return decorator

    def patch(self, path: str, **kwargs):
        """
        Decorator to add a PATCH route.

        Args:
            path (str): The URL path for the route.
            **kwargs: Additional keyword arguments for the route.

        Returns:
            Callable: The decorator function.
        """

        @wraps(self.document("PATCH", path, **kwargs))
        def decorator(func):
            self.router.add_route(
                "PATCH", path, self.document("PATCH", path, **kwargs)(func), **kwargs
            )
            return func

        return decorator

    def delete(self, path: str, **kwargs):
        """
        Decorator to add a DELETE route.

        Args:
            path (str): The URL path for the route.
            **kwargs: Additional keyword arguments for the route.

        Returns:
            Callable: The decorator function.
        """

        @wraps(self.document("DELETE", path, **kwargs))
        def decorator(func):
            self.router.add_route(
                "DELETE", path, self.document("DELETE", path, **kwargs)(func), **kwargs
            )
            return func

        return decorator

    def options(self, path: str, **kwargs):
        """
        Decorator to add a OPTIONS route.
        
        Args:
            path (str): The URL path for the route.
            **kwargs: Additional keyword arguments for the route.
            
        Returns:
            Callable: The decorator function.
        """

        @wraps(self.document("OPTIONS", path, **kwargs))
        def decorator(func):
            self.router.add_route(
                "OPTIONS",
                path,
                self.document("OPTIONS", path, **kwargs)(func),
                **kwargs,
            )
            return func

        return decorator

    def head(self, path: str, **kwargs):
        """
        Decorator to add a HEAD route.
        
        Args:
            path (str): The URL path for the route.
            **kwargs: Additional keyword arguments for the route.
            
        Returns:
            Callable: The decorator function.
        """

        @wraps(self.document("HEAD", path, **kwargs))
        def decorator(func):
            self.router.add_route(
                "HEAD", path, self.document("HEAD", path, **kwargs)(func), **kwargs
            )
            return func

        return decorator

    def trace(self, path: str, **kwargs):
        """
        Decorator to add a TRACE route.
        
        Args:
            path (str): The URL path for the route.
            **kwargs: Additional keyword arguments for the route.
            
        Returns:
            Callable: The decorator function.
        """

        @wraps(self.document("TRACE", path, **kwargs))
        def decorator(func):
            self.router.add_route(
                "TRACE", path, self.document("TRACE", path, **kwargs)(func), **kwargs
            )
            return func

        return decorator

    def connect(self, path: str, **kwargs):
        """
        Decorator to add a CONNECT route.
        
        Args:
            path (str): The URL path for the route.
            **kwargs: Additional keyword arguments for the route.
            
        Returns:
            Callable: The decorator function.
        """

        @wraps(self.document("CONNECT", path, **kwargs))
        def decorator(func):
            self.router.add_route(
                "CONNECT",
                path,
                self.document("CONNECT", path, **kwargs)(func),
                **kwargs,
            )
            return func

        return decorator



"""
---
title: AioFauna
---

# AioFauna

 🚀 Introducing aiofauna: A powerful library to supercharge your FaunaDB experience with modern async Python frameworks! 🔥

🌟 Features:

✅ Async/await coroutines: Leverage the power of async programming for enhanced 
performance and responsiveness.

✅ SSE (Server-Sent Events) support: Stream data in real-time to your clients.

✅ Pydantic-based Document Object Mapper (DOM): Define and validate your data models with ease.

✅ Auto-provisioning: Automatic management of indexes, unique indexes, and collections.

✅ Standardized CRUD operations: Simplify your interactions with FaunaDB using find, find_unique, find_many, find_all, create, upsert, delete, and query methods.

✅ Full JSON communication: Custom encoders to ensure seamless data exchange between your application and FaunaDB backend.

✅ ASGI compliant: `aiofauna.asgi` module provides a middleware to convert `aiohttp.web.Application` into an ASGI application.

💡 With aiofauna, you can build fast, scalable, and reliable applications using the power of FaunaDB and modern asynchronous Python with its out of the box `aiohttp` based web framework. Say goodbye to the hassle of manually managing indexes and collections and hello to a seamless data driven development experience with Pydantic.

🌐 aiofauna is independent and allows native interaction with external services like Docker API, GCP API, AWS API among others, implementing a lightweight stack with aiohttp server capabilities and fauna backend (to be enhanced soon).

📚 Check out the aiofauna library, and start building your next-gen applications today! 🚀
#Python #FaunaDB #Async #Pydantic #aiofauna

⚙️  If you are using a synchronous framework check out [Faudantic](https://github.com/obahamonde/faudantic) for a similar experience with FaunaDB and Pydantic.

📚 [Documentation](https://aiofauna.smartpro.solutions)

📦 [PyPi](https://pypi.org/project/aiofauna/)

📦 [GitHub](https://github.com/obahamonde/aiofauna)

📦 [Demo](https://aiofaunastreams-fwuw7gz7oq-uc.a.run.app/) (Stream data in real-time to your clients) 

"""

__version__ = (0, 0, 14)
__author__ = "obahamonde"
__license__ = "MIT"

import asyncio
from time import perf_counter
from datetime import datetime, date, time, timedelta
from typing import Any, AsyncGenerator, Callable, Dict, List, Optional, Union
from uuid import UUID, uuid4

from aiohttp.web import (
    Application,
    AppRunner,
    Request,
    Response,
    RouteDef,
    TCPSite,
    json_response,
)
from aiohttp.web_middlewares import middleware
from pydantic import (
    BaseModel,
    BaseSettings,
    BaseConfig,
    EmailStr,
    HttpUrl,
    validator,
)  # pylint: disable=no-name-in-module
from aiohttp_sse import sse_response
from aiofauna.client import ClientSession as AioSession, AsyncFaunaClient as AioClient
from aiofauna.errors import AioFaunaException as AioException
from aiofauna import query as q
from aiofauna.json import FaunaJSONEncoder, to_json
from aiofauna.odm import AsyncFaunaModel as AioModel
from aiofauna.application import App as AioApp
from aiofauna.helpers import (
    jsonify,
    redirect,
    render_template,
    make_response,
    upload_files,
)
from aiofauna.asgi import aioasgi
from aiofauna.datastructures import Graph
from aiofauna.api import Api



"""REST API Module with automatic OpenAPI generation."""
import os
import base64
import asyncio
from typing import (
    Any,
    Dict,
)
from inspect import signature
from uuid import UUID
from datetime import datetime, date, time
from enum import Enum
import aiohttp_cors
from aiohttp.web import (
    Application,
    Request,
    Response,
    json_response,
    AppRunner,
    TCPSite,
)
from aiohttp.web_request import FileField
from aiofauna.odm import AsyncFaunaModel as BaseModel
from aiofauna.json import FaunaJSONEncoder as JSONEncoder


def jsonable_encoder(
    obj,
    *,
    include=None,
    exclude=None,
    by_alias=False,
    skip_defaults=False,
    custom_encoder=None,
):
    """
    Convert any object to a JSON-serializable object.

    This function is used by Aiofauna to convert objects to JSON-serializable objects.

    It supports all the types supported by the standard json library, plus:
    
    * datetime.datetime
    * datetime.date
    * datetime.time
    * uuid.UUID
    * enum.Enum
    * pydantic.BaseModel
    
    """

    if custom_encoder is None:
        custom_encoder = JSONEncoder
    if obj == str:
        return "string"
    if obj in (int, float):
        return "number"
    if obj == bool:
        return "boolean"
    if obj == None:
        return "null"
    if obj == list:
        return "array"
    if obj == dict:
        return "object"
    if obj == bytes:
        return "binary"
    if obj == datetime:
        return "date-time"
    if obj == date:
        return "date"
    if obj == time:
        return "time"
    if obj == UUID:
        return "uuid"
    if obj == Enum:
        return "enum"
    if isinstance(obj, (str, int, float, bool, type(None))):
        return obj
    if isinstance(obj, (list, tuple, set, frozenset)):
        return [
            jsonable_encoder(
                v,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                custom_encoder=custom_encoder,
            )
            for v in obj
        ]
    if isinstance(obj, dict):
        return {
            jsonable_encoder(
                k,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                custom_encoder=custom_encoder,
            ): jsonable_encoder(
                v,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                custom_encoder=custom_encoder,
            )
            for k, v in obj.items()
        }
    if isinstance(obj, bytes):
        return base64.b64encode(obj).decode()

    if isinstance(obj, (set, frozenset)):
        return [
            jsonable_encoder(
                v,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                custom_encoder=custom_encoder,
            )
            for v in obj
        ]
    if isinstance(obj, (datetime, date, time)):
        return obj.isoformat()
    if isinstance(obj, Enum):
        return obj.value
    if isinstance(obj, UUID):
        return str(obj)
    return custom_encoder().default(obj)


def extract_parameters_without_request(params: dict):

    """
    
    Helper function to extract parameters from a function without the request parameter.
    
    """

    open_api_params = {}

    for name, param in params.items():
        type_ = param.annotation

        if type_ in (str, int, float, bool):
            open_api_params[name] = {
                "in": "query",
                "name": name,
                "required": True,
                "schema": {"type": type_, "default": param.default, "required": True},
            }

        elif issubclass(type_, BaseModel):
            open_api_params[name] = {
                "in": "body",
                "name": name,
                "required": True,
                "schema": {
                    "type": "object",
                    "default": param.default,
                    "required": True,
                },
            }

        elif issubclass(type_, FileField):
            open_api_params[name] = {
                "in": "formData",
                "name": name,
                "required": True,
                "schema": {"type": "file", "default": param.default, "required": True},
            }

        elif issubclass(type_, Request):
            continue

        else:
            continue

    return {}, open_api_params


def update_open_api(
    open_api: Dict[str, Any],
    path: str,
    method: str,
    func: Any,
    open_api_params: Dict[str, Any],
) -> None:

    """

    Helper function to update the open_api dictionary with the parameters of a function.
    
    """

    open_api["paths"].setdefault(path, {})[method.lower()] = {
        "summary": func.__name__,
        "description": func.__doc__,
        "parameters": list(open_api_params.values()),
        "responses": {"200": {"description": "OK"}},
    }


async def extract_parameters(request: Request, params: dict):

    """
    
    Helper function to extract parameters from a function.
    
    """

    args_to_apply = {}
    open_api_params = {}

    for name, param in params.items():
        type_ = param.annotation

        if type_ in (str, int, float, bool) and name in request.match_info:
            open_api_params[name] = {
                "in": "path",
                "name": name,
                "required": True,
                "schema": {"type": type_, "default": param.default, "required": True},
            }
            args_to_apply[name] = request.match_info[name]

        elif type_ in (str, int, float, bool) and name in request.query:
            open_api_params[name] = {
                "in": "query",
                "name": name,
                "required": True,
                "schema": {"type": type_, "default": param.default, "required": True},
            }
            args_to_apply[name] = type_(request.query[name])

        elif issubclass(type_, BaseModel):
            open_api_params[name] = {
                "in": "body",
                "name": name,
                "required": True,
                "schema": {
                    "type": "object",
                    "default": param.default,
                    "required": True,
                },
            }
            args_to_apply[name] = type_(**await request.json())

        elif issubclass(type_, FileField):
            open_api_params[name] = {
                "in": "formData",
                "name": name,
                "required": True,
                "schema": {"type": "file", "default": param.default, "required": True},
            }
            args_to_apply[name] = await request.post()

        elif issubclass(type_, Request):

            continue

        else:

            continue

    return args_to_apply, open_api_params


class Api(Application):
    """
    
    Api class to create a fastapi-like api.
    
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.openapi = {
            "openapi": "3.0.0",
            "info": {"title": "API", "version": "1.0.0"},
            "paths": {},
            "components": {},
        }
        self._route_open_api_params = {}

        @self.get("/openapi.json")
        async def openapi():
            response = jsonable_encoder(self.openapi)
            return json_response(response)

        @self.get("/docs")
        async def docs():
            html = """<!DOCTYPE html>
            <html lang="en">
            <head>
                <meta charset="UTF-8">
                <title>AioFauna</title>
                <link rel="stylesheet" type="text/css" href="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3.20.3/swagger-ui.css" >
                <link rel="icon" type="image/png" href="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3.20.3/favicon-32x32.png" sizes="32x32" />
                <link rel="icon" type="image/png" href="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3.20.3/favicon-16x16.png" sizes="16x16" />
                <style>
                html
                {
                    box-sizing: border-box;
                    overflow: -moz-scrollbars-vertical;
                    overflow-y: scroll;
                }

                *,
                *:before,
                *:after
                {
                    box-sizing: inherit;
                }

                body
                {
                    margin:0;
                    background: #fafafa;
                }
                </style>
            </head>

            <body>
                <div id="swagger-ui"></div>

                <script src="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3.20.3/swagger-ui-bundle.js"> </script>
                <script src="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3.20.3/swagger-ui-standalone-preset.js"> </script>
                <script>
                window.onload = function() {
                // Begin Swagger UI call region
                const ui = SwaggerUIBundle({
                    url: "/openapi.json",
                    dom_id: '#swagger-ui',
                    deepLinking: true,
                    presets: [
                    SwaggerUIBundle.presets.apis,
                    SwaggerUIStandalonePreset
                    ],
                    plugins: [
                    SwaggerUIBundle.plugins.DownloadUrl
                    ],
                    layout: "StandaloneLayout"
                })
                // End Swagger UI call region

                window.ui = ui
                }
            </script>
            </body>
            </html>
            """
            return Response(body=html.encode(), content_type="text/html")

    def document(self, path: str, method: str):

        """
        
        Decorator to document a function.
        
        """

        def decorator(func):
            sig = signature(func)
            params = sig.parameters
            _, open_api_params = extract_parameters_without_request(params.copy())
            self._route_open_api_params[(path, method)] = open_api_params
            update_open_api(self.openapi, path, method, func, open_api_params)

            async def wrapper(*args, **kwargs):
                request: Request = args[0]
                args = args[1:]
                args_to_apply, _ = await extract_parameters(request, params.copy())
                if asyncio.iscoroutinefunction(func):
                    return await func(*args, **kwargs, **args_to_apply)
                return func(*args, **kwargs, **args_to_apply)

            wrapper._handler = func
            return wrapper

        return decorator

    def get(self, path: str, **kwargs):
        def decorator(func):
            self.router.add_get(path, self.document(path, "GET")(func), **kwargs)
            return func

        return decorator

    def post(self, path: str, **kwargs):
        def decorator(func):
            self.router.add_post(path, self.document(path, "POST")(func), **kwargs)
            return func

        return decorator

    def put(self, path: str, **kwargs):
        def decorator(func):
            self.router.add_put(path, self.document(path, "PUT")(func), **kwargs)
            return func

        return decorator

    def delete(self, path: str, **kwargs):
        def decorator(func):
            self.router.add_delete(path, self.document(path, "DELETE")(func), **kwargs)
            return func

        return decorator

    def patch(self, path: str, **kwargs):
        def decorator(func):
            self.router.add_patch(path, self.document(path, "PATCH")(func), **kwargs)
            return func

        return decorator

    def head(self, path: str, **kwargs):
        def decorator(func):
            self.router.add_head(path, self.document(path, "HEAD")(func), **kwargs)
            return func

        return decorator

    def options(self, path: str, **kwargs):
        def decorator(func):
            self.router.add_options(
                path, self.document(path, "OPTIONS")(func), **kwargs
            )
            return func

        return decorator

    def static(self, path: str, directory: str, **kwargs):
        def decorator(func):
            self.router.add_static(path, directory, **kwargs)
            return func

        return decorator

    async def listen(self, host: str = "0.0.0.0", port: int = 8000, **kwargs):

        cors = aiohttp_cors.setup(
            self,
            defaults={
                "*": aiohttp_cors.ResourceOptions(
                    allow_credentials=True, expose_headers="*", allow_headers="*",
                )
            },
        )
        for route in list(self.router.routes()):
            cors.add(route)
            handler = route.handler._handler  # pylint: disable=protected-access
            path = route.resource.canonical  # type: ignore
            method = route.method.lower()
            open_api_params = self._route_open_api_params.get((path, method), {})
            update_open_api(self.openapi, path, method, handler, open_api_params)

        runner = AppRunner(self)

        await runner.setup()

        site = TCPSite(runner, host, port)

        await site.start()

        print(f"http://{host}:{port}/docs")

        while True:

            await asyncio.sleep(3600)



import warnings
import functools


def deprecated(reason):
    def decorator(old_func):
        @functools.wraps(old_func)
        def new_func(*args, **kvargs):
            fmt = "{name}: {reason}"
            warnings.warn(
                fmt.format(name=old_func.__name__, reason=reason),
                category=DeprecationWarning,
                stacklevel=2,
            )
            return old_func(*args, **kvargs)

        return new_func

    return decorator



"""Lightweight ORM to perform simple CRUD operations on FaunaDB collections and provision indexes

   the fauna query object is available also within the class for further customization
"""

from __future__ import annotations

import asyncio

from typing import List, Optional, Any, Callable

from aiofauna import query as q

from aiofauna.errors import AioFaunaException

from aiofauna.json import JSONModel  # pylint: disable=no-name-in-module

from aiofauna.client import AsyncFaunaClient


class AsyncFaunaModel(JSONModel):
    """

    A base model class for interacting with FaunaDB using asynchronous operations.


    Attributes:
    -----------

    ref: Optional[int]

        The FaunaDB reference ID of the model instance.

    ts: Optional[int]

        The timestamp of the model instance.


    Methods:
    --------

    client() -> AsyncFaunaClient:

        Returns an instance of AsyncFaunaClient to interact with FaunaDB.

    q() -> Expr:

        Returns an instance of Expr to build queries for FaunaDB.

    async provision() -> None:

        Creates a collection and indexes for the model if they don't exist in FaunaDB.

    exists(ref: int) -> bool:

        Checks if a document exists in FaunaDB.

    find_unique(field: str, value: Any) -> Optional[AsyncFaunaModel]:

        Finds a document in FaunaDB by a unique field.

    find_many(field: str, value: Any) -> Optional[List[AsyncFaunaModel]]:

        Finds documents in FaunaDB by a field.

    find(ref: int) -> Optional[AsyncFaunaModel]:

        Finds a document in FaunaDB by its ID.

    find_all() -> Optional[List[AsyncFaunaModel]]:

        Finds all documents of the model in FaunaDB.

    delete_unique(field: str, value: Any) -> bool:

        Deletes a document in FaunaDB by a unique field.

    create() -> Optional[AsyncFaunaModel]:

        Creates a new document in FaunaDB.

    update(ref: int, **kwargs) -> Optional[AsyncFaunaModel]:

        Updates a document in FaunaDB.

    upsert() -> Optional[AsyncFaunaModel]:

        Upserts a document in FaunaDB.

    query(query: str) -> Optional[List[AsyncFaunaModel]]:

        Runs a query in FaunaDB.
    """

    ref: Optional[float] = None

    ts: Optional[float] = None

    def __init__(self, **data: Any) -> None:

        for field in self.__fields__.values():

            try:

                one_of = field.field_info.extra.get("oneOf")

                if isinstance(one_of, list):

                    if data.get(field.name) not in one_of:

                        raise ValueError(f"{field.name} must be one of {one_of}")

            except KeyError:

                continue

        super().__init__(**data)

    @classmethod
    def client(cls) -> AsyncFaunaClient:
        """

        Returns an instance of AsyncFaunaClient to interact with FaunaDB.


        Returns:
        --------

        AsyncFaunaClient:

            An instance of AsyncFaunaClient.
        """

        return AsyncFaunaClient()  # pylint: disable=no-value-for-parameter

    @classmethod
    def q(cls) -> Callable:
        """

        Returns an instance of Expr to build queries for FaunaDB.


        Returns:
        --------

        Expr:

            An instance of Expr.
        """

        return cls.client().query

    @classmethod
    async def provision(cls) -> None:
        """

        Creates a collection and indexes for the model if they don't exist in FaunaDB.


        Returns:
        --------

        None:

            None.
        """

        _q = cls.q()

        try:

            if not await _q(q.exists(q.collection(cls.__name__.lower()))):

                await _q(q.create_collection({"name": cls.__name__.lower()}))

                print(f"Created collection {cls.__name__.lower()}")

                await _q(
                    q.create_index(
                        {
                            "name": cls.__name__.lower(),
                            "source": q.collection(cls.__name__.lower()),
                        }
                    )
                )

                print(f"Created index {cls.__name__.lower()}")

            for field in cls.__fields__.values():

                if field.field_info.extra.get("unique"):

                    await _q(
                        q.create_index(
                            {
                                "name": f"{cls.__name__.lower()}_{field.name}_unique",
                                "source": q.collection(cls.__name__.lower()),
                                "terms": [{"field": ["data", field.name]}],
                                "unique": True,
                            }
                        )
                    )

                    print(
                        f"Created unique index {cls.__name__.lower()}_{field.name}_unique"
                    )
                    continue

                if field.field_info.extra.get("index"):

                    await _q(
                        q.create_index(
                            {
                                "name": f"{cls.__name__.lower()}_{field.name}",
                                "source": q.collection(cls.__name__.lower()),
                                "terms": [{"field": ["data", field.name]}],
                            }
                        )
                    )

                    print(f"Created index {cls.__name__.lower()}_{field.name}")
                    continue

        except AioFaunaException as exc:

            print(exc)

            return None

    @classmethod
    async def exists(cls, ref: float) -> bool:
        """

        Checks if a document exists in FaunaDB.


        Parameters:
        -----------
        ref: int

            The reference ID of the document to check.


        Returns:
        --------

        bool:

            True if the document exists, False otherwise.
        """

        try:

            return await cls.q()(
                q.exists(q.ref(q.collection(cls.__name__.lower()), ref))
            )

        except AioFaunaException as exc:

            print(exc)

            return False

    @classmethod
    async def find_unique(cls, field: str, value: Any) -> Optional[AsyncFaunaModel]:
        """

        Finds a document in FaunaDB by a unique field.


        Parameters:
        -----------
        field: str

            The name of the field to search.

        value: Any

            The value to search for.


        Returns:
        --------

        Optional[AsyncFaunaModel]:

            An instance of the model if found, None otherwise.
        """

        try:

            data = await cls.q()(
                q.get(q.match(q.index(f"{cls.__name__.lower()}_{field}_unique"), value))
            )
            return cls(
                **{
                    **data["data"],
                    "ref": data["ref"]["@ref"]["id"],
                    "ts": data["ts"] / 1000,
                }
            )

        except AioFaunaException as exc:
            print(exc)

            return None

    @classmethod
    async def find_many(cls, field: str, value: Any) -> Optional[List[AsyncFaunaModel]]:
        """

        Finds documents in FaunaDB by a field.


        Parameters:
        -----------
        field: str

            The name of the field to search.

        value: Any

            The value to search for.


        Returns:
        --------

        Optional[List[AsyncFaunaModel]]:

            A list of instances of the model if found, None otherwise.
        """

        try:

            _q = cls.q()

            refs = (
                await _q(
                    q.paginate(
                        q.match(q.index(f"{cls.__name__.lower()}_{field}"), value)
                    )
                )
            )["data"]

            data = await asyncio.gather(
                *[
                    _q(
                        q.get(
                            q.ref(q.collection(cls.__name__.lower()), ref["@ref"]["id"])
                        )
                    )
                    for ref in refs
                ]
            )

            return [
                cls(
                    **{**d["data"], "ref": d["ref"]["@ref"]["id"], "ts": d["ts"] / 1000}
                )
                for d in data
            ]

        except AioFaunaException as exc:

            print(exc)

            return None

    @classmethod
    async def find(cls, ref: float) -> Optional[AsyncFaunaModel]:
        """

        Finds a document in FaunaDB by its ID.


        Parameters:
        -----------
        ref: int

            The reference ID of the document to find.


        Returns:
        --------

        Optional[AsyncFaunaModel]:

            An instance of the model if found, None otherwise.
        """

        try:

            data = await cls.q()(q.get(q.ref(q.collection(cls.__name__.lower()), ref)))
            return cls(
                **{
                    **data["data"],
                    "ref": data["ref"]["@ref"]["id"],
                    "ts": data["ts"] / 1000,
                }
            )

        except AioFaunaException as exc:

            print(exc)

            return None

    @classmethod
    async def find_all(cls) -> Optional[List[AsyncFaunaModel]]:
        """

        Finds all documents of the model in FaunaDB.


        Returns:
        --------

        Optional[List[AsyncFaunaModel]]:

            A list of instances of the model if found, None otherwise.
        """

        try:

            _q = cls.q()

            refs = (await _q(q.paginate(q.match(q.index(cls.__name__.lower())))))[
                "data"
            ]

            data = await asyncio.gather(
                *[
                    _q(
                        q.get(
                            q.ref(q.collection(cls.__name__.lower()), ref["@ref"]["id"])
                        )
                    )
                    for ref in refs
                ]
            )

            return [
                cls(
                    **{**d["data"], "ref": d["ref"]["@ref"]["id"], "ts": d["ts"] / 1000}
                )
                for d in data
            ]

        except AioFaunaException as exc:

            print(exc)

            return None

    @classmethod
    async def delete_unique(cls, field: str, value: Any) -> bool:
        """

        Deletes a document in FaunaDB by a unique field.


        Parameters:
        -----------
        field: str

            The name of the unique field to use in the search.

        value: Any

            The value of the unique field to use in the search.


        Returns:
        --------

        bool:

            True if the document is deleted, False otherwise.
        """

        try:

            _q = cls.q()

            ref = await _q(
                q.get(q.match(q.index(f"{cls.__name__.lower()}_{field}_unique"), value))
            )

            await _q(q.delete(ref))

            return True

        except AioFaunaException as exc:

            print(exc)

            return False

    @classmethod
    async def delete(cls, ref: float) -> bool:

        """Delete a document by id"""

        try:

            return await cls.q()(
                q.delete(q.ref(q.collection(cls.__name__.lower())), ref)
            )

        except AioFaunaException as exc:

            raise AioFaunaException(exc)

    async def create(self) -> Optional[AsyncFaunaModel]:
        """

        Creates a new document in FaunaDB.


        Parameters:
        -----------

        **kwargs:

            The data to create the new document with.


        Returns:
        --------

        Optional[AsyncFaunaModel]:

            An instance of the model if created successfully, None otherwise.
        """

        try:

            for field in self.__fields__.values():

                if field.field_info.extra.get("unique"):

                    instance = await self.find_unique(
                        field.name, self.dict()[field.name]  # type: ignore
                    )
                    if instance:
                        return instance

            data = await self.q()(
                q.create(
                    q.collection(self.__class__.__name__.lower()), {"data": self.dict()}
                )
            )
            return self.__class__(
                **{
                    **data["data"],
                    "ref": data["ref"]["@ref"]["id"],
                    "ts": data["ts"] / 1000,
                }
            )

        except AioFaunaException as exc:

            print(exc)

            return None

    @classmethod
    async def update(cls, ref: float, **kwargs) -> Optional[AsyncFaunaModel]:
        """

        Creates a new document in FaunaDB.


        Parameters:
        -----------

        **kwargs:

            The data to create the new document with.


        Returns:
        --------

        Optional[AsyncFaunaModel]:

            An instance of the model if created successfully, None otherwise.
        """

        try:

            data = await cls.q()(
                q.update(
                    q.ref(q.collection(cls.__name__.lower()), ref), {"data": kwargs}
                )
            )
            return cls(
                **{
                    **data["data"],
                    "ref": data["ref"]["@ref"]["id"],
                    "ts": data["ts"] / 1000,
                }
            )

        except AioFaunaException as exc:

            print(exc)

            return None

    async def upsert(self) -> Optional[AsyncFaunaModel]:
        """

        Creates a new document in FaunaDB.


        Parameters:
        -----------

        **kwargs:

            The data to create the new document with.


        Returns:
        --------

        Optional[AsyncFaunaModel]:

            An instance of the model if created successfully, None otherwise.
        """

        try:
            if not self.ref:

                return await self.create()

            return await self.update(self.ref, **self.dict())  # type: ignore

        except AioFaunaException as exc:

            print(exc)

            return None

    @classmethod
    async def query(cls, query: str) -> Optional[List[AsyncFaunaModel]]:
        """

        Queries FaunaDB using a string query and returns a list of instances of the model.


        Args:

            query (str): The query to use.


        Returns:

            Optional[List[AsyncFaunaModel]]: A list of instances of the model if found, None otherwises.
        """

        try:

            refs = (await cls.q()(q.paginate(q.match(q.query(query)))))["data"]

            data = await asyncio.gather(*[cls.q()(q.get(ref)) for ref in refs])

            return [
                cls(
                    **{**d["data"], "ref": d["ref"]["@ref"]["id"], "ts": d["ts"] / 1000}
                )
                for d in data
            ]

        except AioFaunaException as exc:

            print(exc)

            return None
